@@ -16,7 +16,7 @@ usage () {
     echo "    allegro_s, allegro8_s, allegromodern_s, allegromodern8_s (SMP variants)"
     echo "    allegro_64, allegro8_64, allegromodern_64, allegromodern8_64 (64-bit variants),"
     echo "    allegro_64_s, allegro8_64_s, allegromodern_64_s, allegromodern8_64_s, (SMP, 64-bit variants)"
-    echo "    clisp, cmucl, ecl, gcl, sbcl, scl and xcl."
+    echo "    clasp, clisp, cmucl, ecl, gcl, sbcl, scl and xcl."
     echo " To configure the script, you may set environment variables to point to the various lisp runtimes."
     echo " Allegro CL is a special case: instead of setting environment variables for the specific runtime"
     echo "   locations, you may simply specify the Allegro install directories using these variables:"
@@ -156,7 +156,7 @@ do_tests () {
         echo "or more interactively (and maybe with rlwrap or in emacs), start with:" >&2
         echo "(cd test ; $icmd )" >&2
         echo "then copy/paste:" >&2
-        echo "'(#.(load \"script-support.lisp\") #.(asdf-test::da) #.(load-asdf) #.(frob-packages) #.(asdf::with-asdf-cache () (load \"$i\")))" >&2
+        echo "'(#.(load \"script-support.lisp\") #.(asdf-test::da) #.(load-asdf) #.(frob-packages) #.(load \"$i\"))" >&2
       fi
       echo >&2
       echo >&2
@@ -181,9 +181,9 @@ do_tests () {
 # not used currently but leave here for future reference.
 #
 case $(uname) in
-    CYGWIN*) os=windows ;;
-    Darwin) os=macos ;;
-    Linux) os=linux ;;
+    CYGWIN*|MSYS_NT*) os=windows ;;
+    Darwin*) os=macos ;;
+    Linux*) os=linux ;;
     *) os=unknown ;;
 esac
 
@@ -223,9 +223,9 @@ case "$lisp" in
     # echo ALLEGRO=$ALLEGRO
     flags="-q"
     nodebug="-batch"
-    if [ "$os" = windows ] && [ -z "$ALLEGRO_NOISY" ] ; then
+    if [ "$os" = windows ] ; then
         adir=$(dirname "${command}") ;
-        allegroName=$(basename "${command}") ;
+        allegroName=$(basename "${command}" ".exe") ;
         if [[ ${allegroName: -1} == "8" ]] ; then build=build ; else build=buildi ; fi ;
         # this takes somewhat unjustifiable advantage of the fact that
         # the Allegro images have the same name (with .dxl extension)
@@ -242,8 +242,8 @@ case "$lisp" in
     eval="--eval" ;;
   clasp)
     command="${CLASP:-clasp}"
-    flags=""
-    eval="" ;;
+    flags="--norc --noinit"
+    eval="--eval" ;;
   clisp)
     command="${CLISP:-clisp}"
     flags="-norc --silent -ansi -I "
@@ -284,7 +284,8 @@ case "$lisp" in
     eval="-eval" ;;
   sbcl)
     command="${SBCL:-sbcl}"
-    flags="--noinform --no-userinit --no-sysinit"
+    flags="--no-userinit --no-sysinit"
+    # flags="--noinform --no-userinit --no-sysinit"
     nodebug="--disable-debugger"
     eval="--eval" ;;
   scl)
@@ -307,7 +308,7 @@ if ! type "$command" > /dev/null ; then
     exit 43
 fi
 
-ASDFDIR="$(cd $(dirname $0)/.. ; /bin/pwd)"
+ASDFDIR="$(cd $(dirname $0)/.. ; command pwd)"
 : ${bcmd:=$command $flags} ${icmd:=$command $flags} # batch and interactive
 if [ -z "${DEBUG_ASDF_TEST}" ] ; then
   bcmd="$bcmd $nodebug"
@@ -356,15 +357,27 @@ upgrade_tags () {
     # all the stabilization work done in 3.0 so far, plus extra developments
     # in UIOP, package-inferred-system, and more robustification.
     #   3.1.2 (2014-05-06) is the first ASDF 3.1 release
+    #   3.1.3 (2014-07-24) a bug fix release for 3.1.2
+    #   3.1.4 (2014-10-09) more bug fixes, source-registry cache, in LispWorks 7
+    #   3.1.5 (2015-07-21) more bug fixes, what SBCL sports (as of 1.3.14, 2017-02-04)
+    #   3.1.6 (2015-10-17) more bug fixes
+    #   3.1.7 (2016-03-23) more bug fixes, last in 3.1 series
+    #
+    # The 3.2 series provides the asdf3.2 feature, meaning users can rely on
+    # all its new features (launch-program, improved bundle support), as well as
+    # the improvements done in 3.1 (e.g. XDG support).
+    #   3.2.0 (2017-01-08) first (and latest) in 3.2 series
     #
     # We return the above designated versions in order of decreasing relevance,
     # which pretty much means REQUIRE and most recent first.
-    echo REQUIRE
-    echo 3.1.4 3.1.3 3.1.2
-    echo 3.0.3 3.0.2 3.0.1
-#    echo 2.32 2.27
-#    echo 2.26 2.22 2.20 2.019 2.014.6 2.011 2.008 2.000
-#    echo 1.369 1.97 1.85
+    # We picked the last in each relevant series.
+    echo REQUIRE 3.2.0 3.1.7 3.0.3 2.26
+
+    #echo 3.1.7 3.1.6 3.1.5 3.1.4 3.1.3 3.1.2
+    #echo 3.0.3 3.0.2 3.0.1
+    #echo 2.32 2.27
+    #echo 2.26 2.22 2.20 2.019 2.014.6 2.011 2.008 2.000
+    #echo 1.369 1.97 1.85
 }
 upgrade_methods () {
     if [ -n "$ASDF_UPGRADE_TEST_METHODS" ] ; then
@@ -389,7 +402,7 @@ extract_tagged_asdf () {
         case $tag in
             1.*|2.0*|2.2[0-6]|2.26.61)
                 git show ${tag}:asdf.lisp > $file ;;
-            2.2[7-9]*|2.[3-9]*|3.*)
+            2.2[7-9]*|2.[3-9][0-9]*|3.*)
                 mkdir -p build/old/build
                 git archive ${tag} | (cd build/old/ ; tar xf -)
                 make -C build/old
@@ -413,7 +426,10 @@ valid_upgrade_test_p () {
         # and only need to test it once, below for 2.24.
         abcl:1.*|abcl:2.00[0-9]:*|abcl:201[0-9]:*|abcl:2.2[0-3]:*) : ;;
         # ccl fasl numbering broke loading of old asdf 2.0
-        ccl:2.0[01]*) : ;;
+        ccl:2.0[01]*|ccl:2.2[0-6]*) : ;;
+        # Allegro ships with versions 3*, so give up testing 2
+        # Also, unpatched Allegro 10 has bug updating from 2.26 and before
+        allegro*:[12].*) : ;;
         # My old ubuntu 10.04LTS clisp 2.44.1 came wired in
         # with an antique ASDF 1.374 from CLC that can't be downgraded.
         # More recent CLISPs work.
@@ -501,7 +517,7 @@ test_clean_load () {
       "(or'#.(load(string'|test/script-support.lisp|):verbose():print())#.(asdf-test:exit-lisp'0))" \
       > $nop 2>&1
     $bcmd $eval \
-      "(or'#.(load(string'|test/script-support.lisp|):verbose():print())#.(asdf-test:verbose())#.(load(string'|build/asdf.lisp|):verbose())#.(asdf/image:quit'0))" \
+      "(or'#.(load(string'|test/script-support.lisp|):verbose():print())#.(asdf-test:verbose())#.(load(string'|build/asdf.lisp|):verbose())#.(uiop/image:quit'0))" \
       > $load 2>&1
     if diff $nop $load ; then
       echo "GOOD: Loading ASDF on $lisp produces no message" >&2 ; return 0
